""" This module contains a variety of integration rules.
"""

# SciPy Stack
import scipy.integrate as integrate

import numpy as np

# project library
from testing.checks import basic_checks


def naive_monte_carlo(func, bounds, num_draws, implementation, seed):
    """ This function performs a Monte Carlo integration.
    """
    def fast_implementation(deviates):
        """ This function applies the utility evaluation to all deviates at
        once.
        """
        rslt = np.sum(func(deviates))
        return rslt

    def slow_implementation(deviates):
        """ This function applies the utility evaluation one deviate at a time.
        """
        rslt = 0.0
        for deviate in deviates:
            rslt += func(deviate)
        return rslt

    # Guard interface.
    args = (func, bounds, num_draws, implementation, seed)
    assert basic_checks('naive_monte_carlo', 'in', args)

    # Distribute bounds.
    lower, upper = bounds

    # Draw requested number of deviates.
    np.random.seed(seed)
    deviates = np.random.uniform(lower, upper, size=num_draws)

    # Implement native Monte Carlo approach.
    if implementation == 'slow':
        rslt = slow_implementation(deviates)
    elif implementation == 'fast':
        rslt = fast_implementation(deviates)
    else:
        raise AssertionError

    # Scaling by augmented assignment.
    rslt /= num_draws

    # Check result.
    assert basic_checks('naive_monte_carlo', 'out', rslt)

    # Finishing
    return rslt


def scipy_quad(func, bounds):
    """ This function is a  wrapper to the general-purpose quadrature
    function from the SciPy library.
    """
    # Guard interface.
    assert basic_checks('scipy_quad', 'in', func, bounds)

    # Distribute bounds.
    lower, upper = bounds

    # Apply integration.
    rslt = integrate.quad(func, lower, upper)[0]

    # Check result.
    assert basic_checks('scipy_quad', 'out', rslt)

    # Finishing
    return rslt


def scipy_romberg(func, bounds):
    """ This function is a wrapper for the Romberg integration provided in
    the SciPy library.
    """
    # Guard interface.
    assert basic_checks('scipy_romberg', 'in', func, bounds)

    # Distribute bounds.
    lower, upper = bounds

    # Apply integration.
    rslt = integrate.romberg(func, lower, upper)

    # Check result.
    assert basic_checks('scipy_romberg', 'out', rslt)

    # Finishing
    return rslt
